#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2023 Yang
# ararpy - calc - corr
# ==========================================
#
#
#
"""
import traceback

from . import arr, err, basic
import numpy as np
from typing import Tuple, List, Optional, Union
from types import MethodType
from math import exp, log


def blank(a0: list, e0: list, a1: list, e1: list):
    """
    :param a0: a list of tested isotope value
    :param e0: 1 sigma error of a0, list type
    :param a1: a list of blank isotope value
    :param e1: 1 sigma error of a1, list type
    :return: list of corrected data | error list
    """
    # Do not force negative value to zero in correcting blank...
    return arr.sub((a0, e0), (a1, e1))


def gain(a0: list, e0: list, a1: list, e1: list):
    """
    :param a0: a list of tested isotope value
    :param e0: 1 sigma error of a0, list type
    :param a1: a list of blank isotope value
    :param e1: 1 sigma error of a1, list type
    :return: list of corrected data | error list
    """
    # Do not force negative value to zero in correcting blank...
    return arr.div((a0, e0), (a1, e1))


def mdf(rm: float, srm: float, m1: float, m2: float, ra: float = 298.56,
        sra: float = 0):
    """
    :param rm: ratio 40a/36a
    :param srm: error of ratio in one sigma
    :param m1: Ar36 isotopic mass
    :param m2: Ar40 isotopic mass
    :param ra: theoretical 40a/36a
    :param sra: error of theoretical 40a/36a
    :return: linear mdf, error, exp mdf, error, pow mdf, error
    """
    sm1 = 0
    sm2 = 0
    delta_m = m2 - m1
    sdelta_m = err.add(sm2, sm1)
    ratio_m = m2 / m1
    sratio_m = err.div((m2, sm2), (m1, sm1))
    useRyu = False
    if not useRyu:
        # line, D = 1 + (ra / rm - 1) / dm  # Renne2009
        k1 = 1 + (ra / rm - 1) / delta_m
        k2 = err.div(((ra / rm - 1), err.div((ra, sra), (rm, srm))), (delta_m, sdelta_m))
        # exp, D = 1 + ln(ra / rm) / (m1 * ln(m2/m1))  # Renne2009
        try:
            k3 = (np.log(ra / rm) / np.log(ratio_m)) * (1 / m1) + 1
            v1 = err.log((ra / rm, err.div((ra, sra), (rm, srm))))
            v2 = err.log((ratio_m, sratio_m))
            v3 = err.div((np.log(ra / rm), v1), (np.log(ratio_m), v2))
            k4 = err.div((np.log(ra / rm) / np.log(ratio_m), v3), (m1, sm1))
        except Exception:
            k3, k4 = 0, 0
        # pow, D = pow(ra / rm, 1 / dm)  # Renne2009
        try:
            k5 = pow((ra / rm), (1 / delta_m))  # Renne2009, B.D. Turrin2010
            k6 = err.pow((ra / rm, err.div((ra, sra), (rm, srm))),
                         (1 / delta_m, err.div((1, 0), (delta_m, sdelta_m))))
        except Exception:
            k5, k6 = 0, 0
        return k1, k2, k2, k2, k3, k4, k4, k4, k5, k6, k6, k6
    else:
        mdf_line_2 = (rm / ra - 1) / delta_m  # Ryu et al., 2013
        return mdf_line_2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0


def discr(a0: list, e0: list, mdf: list, smdf: list, m: list, m40: list,
          isRelative=True, method="l"):
    """
    :param a0: a list of tested isotope value
    :param e0: 1 sigma error of a0, list type
    :param mdf: mass discrimination factor(MDF), list
    :param smdf: absolute error of MDF, list
    :param m: mass of isotope being corrected
    :param m40: mass of Ar40, default value is defined above
    :param isRelative: errors of params are in a relative format
    :param method: correction method, "l" or "linear", "e" or "exponential", "p" or "power"
    :return: corrected value | error of corrected value
    linear correction, MDF = [(Ar40/Ar36)true / (Ar40/Ar36)measure] * 1 / MD - 1 / MD + 1
    corr = blank_corrected / [MD * MDF - MD +1]
    """
    r0, r1 = [], []
    if isRelative:
        smdf = [smdf[i] * mdf[i] / 100 for i in range(len(smdf))]
    for i in range(min([len(arg) for arg in [a0, e0, mdf, smdf]])):
        delta_m = abs(m40[i] - m[i])
        ratio_m = abs(m40[i] / m[i]) if m[i] != 0 else 1
        if method.lower().startswith("l"):  # linear
            k0 = 1 / (delta_m * mdf[i] - delta_m + 1) if (delta_m * mdf[i] - delta_m + 1) != 0 else 0
            k1 = err.rec((delta_m * mdf[i] - delta_m + 1, smdf[i] * delta_m))
        elif method.lower().startswith("e"):  # exponential
            k0 = 1 / np.exp((mdf - 1) * m * np.log(m40 / m))
            k1 = err.rec((1 / k0, err.pow((np.e, 0), ((mdf - 1) * m * np.log(m40 / m), err.mul((mdf - 1, smdf), (m * np.log(m40 / m), 0))))))
        elif method.lower().startswith("p"):  # power
            k0 = 1 / (mdf[i] ** delta_m)
            k1 = err.rec((mdf[i] ** delta_m, err.pow((mdf[i], smdf[i]), (delta_m, 0))))
        else:
            k0 = 1
            k1 = 0
        r0.append(a0[i] * k0)
        r1.append(err.mul((a0[i], e0[i]), (k0, k1)))
    return [r0, r1]


def decay(a0: list, e0: list, t1: list, t2: list, t3: list, f: list, sf: list,
          unit: str = 'h', isRelative=True):
    """
    Parameters
    ----------
    a0
    e0
    t1 : [year, month, day, hour, min, (second)], test start time
    t2
    t3
    f
    sf
    unit
    isRelative

    Returns
    -------

    """
    r0, r1 = [], []
    if isRelative:
        sf = [sf[i] * f[i] / 100 for i in range(len(sf))]
    for i in range(len(a0)):
        k = get_decay_factor(t1[i], t2[i], t3[i], f[i], sf[i], unit)
        r0.append(a0[i] * k[0])
        r1.append(err.mul((a0[i], e0[i]), (k[0], k[1])))
    return [r0, r1]


def get_decay_factor(t1: list, t2: list, t3: list, f: float, sf: float,
                     unit: str = 'h'):
    """
    :param t1: [year, month, day, hour, min, (second)], test start time
    :param t2: irradiation end time for all cycles, [[year, month, day, hour, min],...]
    :param t3: irradiation durations for all cycles, list for all irradiation cycles, in hour
    :param f: decay constant of K
    :param sf: absolute error of f
    :param unit: unit of decay constant, input 'h' or 'a'
    :return: correction factor | error of factor | stand duration
    """
    v1 = []
    v2 = []
    e1 = []
    # t_year, t_month, t_day, t_hour, t_min = t1
    t_test_start = basic.get_datetime(*t1)  # the time when analysis began
    t2 = [basic.get_datetime(*i) for i in t2]  # the time when irradiation ended for all cycles, in second
    k2 = [t_test_start - i for i in t2]  # standing time in second between irradiation and analysing

    try:
        if unit == 'h':
            k2 = [float(i) / 3600 for i in k2]  # exchange to unit in hour
            t3 = [float(i) for i in t3]
        elif unit == 'a':
            k2 = [i / (3600 * 24 * 365.242) for i in k2]  # exchange to unit in year
            t3 = [float(i) / (24 * 365) for i in t3]
        for i in range(len(t3)):
            iP = 1  # power
            v1.append(iP * (1 - np.exp(-f * t3[i])) / (f * np.exp(f * k2[i])))
            e11 = t3[i] * np.exp(-f * t3[i]) / (f * np.exp(f * k2[i]))
            e12 = (np.exp(-f * t3[i]) - 1) * (1 + f * k2[i]) * np.exp(f * k2[i]) / (f * np.exp(f * k2[i])) ** 2
            e1.append(iP * (e11 + e12))
            v2.append(iP * t3[i])
        k0 = sum(v2) / sum(v1)
        k1 = err.div((sum(v2), 0), (sum(v1), pow(sum(e1) ** 2 * sf ** 2, 0.5)))
        # other error calculation equation in CALC
        # It is calculated based on an assumption that only one irradiation exist with total duration of sum of t3,
        # and the end time is the last irradiation finish time
        k1 = pow(
            ((sum(t3) * np.exp(f * k2[-1]) * (1 - np.exp(-f * sum(t3))) + f * sum(t3) * k2[-1] * np.exp(f * k2[-1]) * (
                    1 - np.exp(-f * sum(t3))) - f * sum(t3) * np.exp(f * k2[-1]) * sum(t3) * np.exp(-f * sum(t3))) / (
                     1 - np.exp(-f * sum(t3))) ** 2) ** 2 * sf ** 2, 0.5)
    except Exception as e:
        print(traceback.format_exc())
        return 1, 0
    else:
        return k0, k1


def get_irradiation_datetime_by_string(datetime_str: str):
    """
    Parameters
    ----------
    datetime_str : datatime string, like "2022-04-19-18-35D13.45S2022-04-19-04-19D6.7"
        return [2022-04-19T18:35:13.45, 2022-04-19T04:19:6.7]

    Returns
    -------

    """
    res = []
    if datetime_str == '' or datetime_str is None or (isinstance(datetime_str, float) and np.isnan(datetime_str)):
        return ['', 0]
    cycles = datetime_str.split('S')
    for cycle in cycles:
        [dt, hrs] = cycle.split('D')
        res = res + [dt, float(hrs)]
    return res


def get_method_fitting_law_by_name(method_str: str):
    """
    Parameters
    ----------
    method_str

    Returns
    -------

    """
    res = [False] * 3
    try:
        res[['Lin', 'Exp', 'Pow'].index(method_str.capitalize()[:3])] = True
    except (ValueError, AttributeError):
        pass
    return res


def random_normal_relative(mu, std, size):
    return np.random.normal(mu, std * mu / 100, size)


def random_normal_absolute(mu, std, size):
    if np.isnan(std):
        std = 0
    return np.random.normal(mu, std, size)


def Monte_Carlo_F(ar40m: Tuple[float, float], ar39m: Tuple[float, float], ar38m: Tuple[float, float],
                  ar37m: Tuple[float, float], ar36m: Tuple[float, float],
                  ar40b: Tuple[float, float], ar39b: Tuple[float, float], ar38b: Tuple[float, float],
                  ar37b: Tuple[float, float], ar36b: Tuple[float, float],
                  t1: List[float], t2: List[List[float]], t3: List[float],
                  R40v36a: Tuple[float, float], R38v36a: Tuple[float, float],
                  R39v37ca: Tuple[float, float], R36v37ca: Tuple[float, float], R38v37ca: Tuple[float, float],
                  R40v39k: Tuple[float, float], R38v39k: Tuple[float, float],
                  R36v38clp: Tuple[float, float],
                  L37ar: Tuple[float, float], L39ar: Tuple[float, float], L36cl: Tuple[float, float],
                  MDFunc: Union[MethodType],
                  MDF: Tuple[float, float],
                  M40: Tuple[float, float], M39: Tuple[float, float], M38: Tuple[float, float],
                  M37: Tuple[float, float], M36: Tuple[float, float],
                  G40: Tuple[float, float], G39: Tuple[float, float], G38: Tuple[float, float],
                  G37: Tuple[float, float], G36: Tuple[float, float],
                  stand_time_year: float, JNFactor: Tuple[float, float],
                  **options):
    """

    Parameters
    ----------
    ar40m
    ar39m
    ar38m
    ar37m
    ar36m
    t1
    t2
    t3
    R40v36a
    R38v36a
    R39v37ca
    R36v37ca
    R40v39k
    R38v39k
    R36v38clp
    L37ar
    L39ar
    L36cl
    MDFunc
    MDF
    JNFactor
    options

    Returns
    -------

    """
    # define parameters

    MDFunc = lambda _M40, _mass, _mdf: 1 / ((_M40 - _mass) * (_mdf - 1) + 1)

    Monte_Carlo_Size = 10000

    # generate random
    ar40m = random_normal_absolute(*ar40m, size=Monte_Carlo_Size)
    ar39m = random_normal_absolute(*ar39m, size=Monte_Carlo_Size)
    ar38m = random_normal_absolute(*ar38m, size=Monte_Carlo_Size)
    ar37m = random_normal_absolute(*ar37m, size=Monte_Carlo_Size)
    ar36m = random_normal_absolute(*ar36m, size=Monte_Carlo_Size)
    ar40b = random_normal_absolute(*ar40b, size=Monte_Carlo_Size)
    ar39b = random_normal_absolute(*ar39b, size=Monte_Carlo_Size)
    ar38b = random_normal_absolute(*ar38b, size=Monte_Carlo_Size)
    ar37b = random_normal_absolute(*ar37b, size=Monte_Carlo_Size)
    ar36b = random_normal_absolute(*ar36b, size=Monte_Carlo_Size)

    MDF =random_normal_relative(*MDF, size=Monte_Carlo_Size)

    M40 = random_normal_relative(*M40, size=Monte_Carlo_Size)
    M39 = random_normal_relative(*M39, size=Monte_Carlo_Size)
    M38 = random_normal_relative(*M38, size=Monte_Carlo_Size)
    M37 = random_normal_relative(*M37, size=Monte_Carlo_Size)
    M36 = random_normal_relative(*M36, size=Monte_Carlo_Size)

    G40 = random_normal_relative(*G40, size=Monte_Carlo_Size)
    G39 = random_normal_relative(*G39, size=Monte_Carlo_Size)
    G38 = random_normal_relative(*G38, size=Monte_Carlo_Size)
    G37 = random_normal_relative(*G37, size=Monte_Carlo_Size)
    G36 = random_normal_relative(*G36, size=Monte_Carlo_Size)

    L36cl = random_normal_relative(*L36cl, size=Monte_Carlo_Size)
    L37ar = random_normal_relative(*L37ar, size=Monte_Carlo_Size)
    L39ar = random_normal_relative(*L39ar, size=Monte_Carlo_Size)

    R40v36a = random_normal_relative(*R40v36a, size=Monte_Carlo_Size)
    R38v36a = random_normal_relative(*R38v36a, size=Monte_Carlo_Size)
    R39v37ca = random_normal_relative(*R39v37ca, size=Monte_Carlo_Size)
    R38v37ca = random_normal_relative(*R38v37ca, size=Monte_Carlo_Size)
    R36v37ca = random_normal_relative(*R36v37ca, size=Monte_Carlo_Size)
    R40v39k = random_normal_relative(*R40v39k, size=Monte_Carlo_Size)
    R38v39k = random_normal_relative(*R38v39k, size=Monte_Carlo_Size)
    R36v38clp = random_normal_relative(*R36v38clp, size=Monte_Carlo_Size)

    JNFactor = random_normal_absolute(*JNFactor, size=Monte_Carlo_Size)

    blank_gain_corr = options.get('blank_gain_corr', True)

    def do_simulation():
        i = 0
        while i < Monte_Carlo_Size:
            P40Mdf = MDFunc(M40[i], M40[i], MDF[i])
            P39Mdf = MDFunc(M40[i], M39[i], MDF[i])
            P38Mdf = MDFunc(M40[i], M38[i], MDF[i])
            P37Mdf = MDFunc(M40[i], M37[i], MDF[i])
            P36Mdf = MDFunc(M40[i], M36[i], MDF[i])
            P37Decay = get_decay_factor(t1, t2, t3, L37ar[i], sf=0)[0]
            P39Decay = get_decay_factor(t1, t2, t3, L39ar[i], sf=0)[0]

            _ar40m = ar40m[i]

            _ar36 = (ar36m[i] / G36[i] - ar36b[i] / (G36[i] if blank_gain_corr else 1)) * P36Mdf
            _ar37 = (ar37m[i] / G37[i] - ar37b[i] / (G37[i] if blank_gain_corr else 1)) * P37Mdf
            _ar38 = (ar38m[i] / G38[i] - ar38b[i] / (G38[i] if blank_gain_corr else 1)) * P38Mdf
            _ar39 = (ar39m[i] / G39[i] - ar39b[i] / (G39[i] if blank_gain_corr else 1)) * P39Mdf
            _ar40 = (ar40m[i] / G40[i] - ar40b[i] / (G40[i] if blank_gain_corr else 1)) * P40Mdf

            _ar37ca = _ar37 * P37Decay
            _ar39k = _ar39 * P39Decay - _ar37ca * R39v37ca[i]
            _ar38res = _ar38 - _ar39k * R38v39k[i] - _ar37ca * R38v37ca[i]
            _ar36res = _ar36 - _ar37ca * R36v37ca[i]
            _ar36cl = (_ar36res - _ar38res / R38v36a[i]) / (1 - 1 / (R36v38clp[i] * (1 - exp(-1 * L36cl[i] * stand_time_year)) * R38v36a[i]))
            _ar36a = _ar36res - _ar36cl
            _ar40r = _ar40 - _ar36a * R40v36a[i] - _ar39k * R40v39k[i]
            _ar40ar = _ar40 - _ar39k * R40v39k[i]
            _ar39k = _ar39k * JNFactor[i]
            _f = _ar40r / _ar39k
            i += 1
            # yield _f, _ar36, _ar37, _ar38, _ar39, _ar40, _ar36a, _ar37ca, _ar39k, _ar40r, _ar36cl
            yield _f, _ar36a, _ar39k, _ar40ar

    simulation = do_simulation()
    F = []
    ar36 = []
    ar37 = []
    ar38 = []
    ar39 = []
    ar40 = []

    ar36a = []
    ar37ca = []
    ar39k = []
    ar40r = []
    ar36cl = []

    nor_x = []
    nor_y = []
    inv_x = []
    inv_y = []
    for each in simulation:
        F.append(each[0])
        nor_x.append(each[2] / each[1])
        nor_y.append(each[3] / each[1])
        inv_x.append(each[2] / each[3])
        inv_y.append(each[1] / each[3])
        # ar36.append(each[1])
        # ar37.append(each[2])
        # ar38.append(each[3])
        # ar39.append(each[4])
        # ar40.append(each[5])
        # ar36a.append(each[6])
        # ar37ca.append(each[7])
        # ar39k.append(each[8])
        # ar40r.append(each[9])
        # ar36cl.append(each[10])

    print(f"{np.mean(F) = }, {np.std(F) = }")

    # print("F = {0} ± {1}".format(np.mean(F), np.std(F)))
    #
    # with open("ar36m.txt", 'w') as f:  # save serialized json data to a readable text
    #     f.writelines("\n".join([str(i) for i in ar36m]))
    # with open("ar37m.txt", 'w') as f:  # save serialized json data to a readable text
    #     f.writelines("\n".join([str(i) for i in ar37m]))
    # with open("ar38m.txt", 'w') as f:  # save serialized json data to a readable text
    #     f.writelines("\n".join([str(i) for i in ar38m]))
    # with open("ar39m.txt", 'w') as f:  # save serialized json data to a readable text
    #     f.writelines("\n".join([str(i) for i in ar39m]))
    # with open("ar40m.txt", 'w') as f:  # save serialized json data to a readable text
    #     f.writelines("\n".join([str(i) for i in ar40m]))
    #
    # with open("ar36.txt", 'w') as f:  # save serialized json data to a readable text
    #     f.writelines("\n".join([str(i) for i in ar36]))
    # with open("ar37.txt", 'w') as f:  # save serialized json data to a readable text
    #     f.writelines("\n".join([str(i) for i in ar37]))
    # with open("ar38.txt", 'w') as f:  # save serialized json data to a readable text
    #     f.writelines("\n".join([str(i) for i in ar38]))
    # with open("ar39.txt", 'w') as f:  # save serialized json data to a readable text
    #     f.writelines("\n".join([str(i) for i in ar39]))
    # with open("ar40.txt", 'w') as f:  # save serialized json data to a readable text
    #     f.writelines("\n".join([str(i) for i in ar40]))
    #
    # with open("ar36a.txt", 'w') as f:  # save serialized json data to a readable text
    #     f.writelines("\n".join([str(i) for i in ar36a]))
    # with open("ar36cl.txt", 'w') as f:  # save serialized json data to a readable text
    #     f.writelines("\n".join([str(i) for i in ar36cl]))
    # with open("ar37ca.txt", 'w') as f:  # save serialized json data to a readable text
    #     f.writelines("\n".join([str(i) for i in ar37ca]))
    # with open("ar39k.txt", 'w') as f:  # save serialized json data to a readable text
    #     f.writelines("\n".join([str(i) for i in ar39k]))
    # with open("ar40r.txt", 'w') as f:  # save serialized json data to a readable text
    #     f.writelines("\n".join([str(i) for i in ar40r]))
    # F, sF,
    return np.mean(F), np.std(F), \
           np.mean(nor_x), np.std(nor_x), np.mean(nor_y), np.std(nor_y), \
           np.cov(nor_x, nor_y)[0][1] / (np.std(nor_x) * np.std(nor_y)), \
           np.mean(inv_x), np.std(inv_x), np.mean(inv_y), np.std(inv_y), \
           np.cov(inv_x, inv_y)[0][1] / (np.std(inv_x) * np.std(inv_y))








"""
# ==========================================
# Copyright 2023 Yang
# ararpy - files - calc_file
# ==========================================

Open age files from ArArCALC. It currently supports version: 25.2 and 24.0.
Version 25.2 is preferred to 24.0 as containing more detailed parameters than 24.0.

ArArCalcFile(file_path, file_name: optional)
    -> open()
        -> get content and sample info ()

"""
import traceback
import pandas as pd
import numpy as np
import re
from xlrd import open_workbook, biffh
import os
import msoffcrypto

from . import xls
from ..calc.basic import get_datetime, parser
from ..calc import arr, err, isochron


def read_calc_file(file_path: str):
    """
    :param file_path: file with suffix of .age or .air
    :return: dict, keys are sheet name, value are list of [[col 0], [col 1], ...] of corresponding sheet
    """
    decrypt_file_path = file_path + '_decrypt.xls'
    try:
        with open(file_path, 'rb') as file:
            office_file = msoffcrypto.OfficeFile(file)
            office_file.load_key(password=bytes.fromhex('6161706b6f70').decode())
            office_file.decrypt(open(decrypt_file_path, 'wb'))
        wb = open_workbook(decrypt_file_path)
        worksheets = wb.sheet_names()
        book_contents = dict()
        for each_sheet in worksheets:
            sheet = wb.sheet_by_name(each_sheet)
            sheet_contents = [
                [sheet.cell(row, col).value for col in range(sheet.ncols)]
                for row in range(sheet.nrows)
            ]
            book_contents[each_sheet] = sheet_contents
        os.remove(decrypt_file_path)
    except Exception as e:
        raise
    else:
        return book_contents


def open_252(data: pd.DataFrame, logs01: pd.DataFrame, logs02: pd.DataFrame):
    """
    Open 25.2 version ArArCALC files
    :param data:
    :param logs01:
    :param logs02:
    :return:
    """
    # add default columns for 999, -999, and -1 index
    data[-999] = [0] * data.index.size
    data[999] = [1] * data.index.size
    data[-1] = [np.nan] * data.index.size
    # rows_unm --> 样品阶段数
    sequence_index = [1, 2, ]
    sample_values_index = [16, 17, 21, 22, 26, 27, 31, 32, 36, 37,]
    blank_values_index = [3, 4, 5, 6, 7, 8, 9, 10, 11, 12,]
    corrected_values_index = [
        188, 189, 190, 191, 192, 193, 194, 195, 196, 197,]
    degas_values_index = [
        138, 139, 140, 141, 142, 143, 144, 145,  # 36 a, c, ca, cl    0-9
        146, 147,  # 37 ca    10-11
        156, 157, 148, 149, 150, 151, 152, 153, 154, 155,  # 38 cl, a, c, k, ca   12-21
        158, 159, 160, 161,  # 39 k, ca    22-25
        162, 163, 164, 165, 166, 167, 168, 169,  # 40 r, a, c, k    26-33
    ]
    publish_values_index = [
        101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111,
    ]
    apparent_age_values_index = [
        198, 199, 200, 201, -999, -999, 202, 203,  # f, sf, ages, s, s, s, 40Arr%, 39Ar%
    ]
    isochron_values_index = [
        116, 117, 118, 119, 122,  # normal isochron
        -1, 127, 128, 129, 130, 133,  # inverse
        -1, -999, -999, -999, -999, -999,  # 39/38 vs 40/38
        -1, -999, -999, -999, -999, -999,  # 39/40 vs 39/40
        -1, -999, -999, -999, -999, -999,  # 38/39 vs 40/39
        -1, -999, -999, -999, -999, -999, -999, -999, -999, -999,  # 3D isochron, 36/39, 38/39, 40/39
    ]
    isochron_mark_index = [115,]
    total_param_index = [
        71, 72, 73, 74, 75, 76, 77, 78, 79, 80,  # 0-9
        81, 82, 83, 84, 85, 86, 87, 88, 89, 90,  # 10-19
        91, 92, 93, 94, 95, 96, -999, -999, 63, -999,  # 20-29
        -999, -999, -1, -1, -999, -999, -999, -999, -999, -999,  # 30-39
        -999, -999, -999, -999, -999, -999, -999, -999, -999, -999,  # 40-49
        -999, -999, -999, -999, -999, -999, -999, -999, 67, 49,  # 50-59
        50, -999, -999, -999, -999, -999, -999, 51, 52, 53,  # 60-69
        54, -999, -999, -999, -999, -999, -999, -999, -999, -999,  # 70-79
        -999, -999, -999, -999, -999, -999, -999, -999, -999, -999,  # 80-89
        -999, -999, -999, -999, -999, -999, -999, -999, -999, -999,  # 90-99
        -999, -999, -999, -999, -999, -999, -999, -999, -999, -999,  # 100-109
        -999, -999, -999, -999, -1,  # 110-114
        -999, -999, -999, -999, -999, -999, -999, -999, -999, -999, -999,  # 115-125
        999, -999, 999, -999, 999, -999, 999, -999, 999, -999,  # 126-135
    ]

    # double transpose to remove keys
    get_data = lambda df, index: pd.concat([df[index].transpose()], ignore_index=True).transpose()

    sample_values = get_data(data, sample_values_index)
    blank_values = get_data(data, blank_values_index)
    corrected_values = get_data(data, corrected_values_index)
    degas_values = get_data(data, degas_values_index)
    publish_values = get_data(data, publish_values_index)
    apparent_age_values = get_data(data, apparent_age_values_index)
    isochron_values = get_data(data, isochron_values_index)
    total_param = get_data(data, total_param_index)
    isochron_mark = get_data(data, isochron_mark_index)
    sequence = get_data(data, sequence_index)

    # adjustment
    isochron_mark.replace({4.0: '1'}, inplace=True)

    total_param[71] = 35.96754628
    total_param[73] = 36.9667759
    total_param[75] = 37.9627322
    total_param[77] = 38.964313
    total_param[79] = 39.962383123

    total_param[83] = logs01[1][9]  # No
    total_param[84] = logs01[1][10]  # %SD
    total_param[81] = logs01[1][11]  # K mass
    total_param[82] = logs01[1][12]  # K mass %SD

    total_param[85] = logs01[1][13]  # How many seconds in a year
    total_param[86] = 0  # %SD

    total_param[87] = logs01[1][14]  # The constant of 40K/K
    total_param[88] = logs01[1][15]  # %SD

    total_param[89] = logs01[1][16]  # The constant of 35Cl/37Cl
    total_param[90] = logs01[1][17]  # %SD
    total_param[56] = logs01[1][18]  # The constant of 36/38Cl productivity
    total_param[57] = logs01[1][19]  # %SD
    total_param[91] = logs01[1][24]  # The constant of HCl/Cl
    total_param[92] = logs01[1][25]  # %SD

    total_param[50] = logs01[1][26]  # 40K(EC) activities param
    total_param[51] = logs01[1][27]  # %SD
    total_param[52] = logs01[1][28]  # 40K(\beta-) activities param
    total_param[53] = logs01[1][29]  # %SD
    total_param[48] = logs01[1][26] + logs01[1][28]  # 40K total activities param
    total_param[49] = round(100 / total_param[48] * pow(
        (logs01[1][27] / 100 * logs01[1][26]) ** 2 + (logs01[1][29] / 100 * logs01[1][28]) ** 2, 0.5), 6)  # %SD

    total_param[34] = logs01[1][36]  # decay constant of 40K total
    total_param[35] = logs01[1][37]  # %SD
    total_param[36] = logs01[1][30]  # decay constant of 40K EC
    total_param[37] = logs01[1][31]  # %SD
    total_param[38] = logs01[1][32]  # decay constant of 40K \beta-
    total_param[39] = logs01[1][33]  # %SD
    total_param[40] = 0  # decay constant of 40K \beta+
    total_param[41] = 0  # %SD
    total_param[46] = logs01[1][34]  # decay constant of 36Cl
    total_param[47] = logs01[1][35]  # %SD
    total_param[42] = logs01[1][38]  # decay constant of 39Ar
    total_param[43] = logs01[1][39]  # %SD
    total_param[44] = logs01[1][40]  # decay constant of 37Ar
    total_param[45] = logs01[1][41]  # %SD

    # logs01[1][49]  # 1 for using weighted YORK isochron regression, 2 for unweighted
    # logs01[1][50]  # True for including errors on irradiation constants, False for excluding those
    # logs01[1][67]  # MDF method, 'LIN'
    total_param[93] = logs01[1][70]  # 40Ar/36Ar air
    total_param[94] = logs01[1][71]  # %SD
    total_param[97] = 'York-2' if int(logs01[1][49]) == 1 else ""   # Fitting method
    total_param[98] = logs01[1][0]  # Convergence
    total_param[99] = logs01[1][1]  # Iterations number
    total_param[100] = logs01[1][67]  # MDF method
    total_param[101] = parser(logs01[1][65], bool)  # force negative to zero when correct blank
    total_param[102] = True  # Apply 37Ar decay
    total_param[103] = True  # Apply 39Ar decay
    total_param[104] = parser(logs01[1][6], bool)  # Apply 37Ar decay
    total_param[105] = parser(logs01[1][7], bool)  # Apply 39Ar decay
    total_param[106] = True  # Apply Ca degas
    total_param[107] = True  # Apply K degas
    total_param[108] = parser(logs01[1][8], bool)  # Apply Cl degas
    total_param[109] = True  # Apply Trap degas

    total_param[110] = True if logs01[1][5] == 'MIN' else False  # Calculating ages using Min equation [True] or conventional equation [False]
    total_param[111] = parser(logs01[1][89], bool)  # Use primary standard or not
    total_param[112] = parser(logs01[1][91], bool)  # Use standard age or not
    total_param[113] = parser(logs01[1][92], bool)  # Use primary ratio or not

    # irradiation time
    total_param = general_adjustment(
        total_param.copy(), logs02, get_data(data, [59, 58, 57, 60, 61]), data[63][0]
    )
    return [
        sample_values, blank_values, corrected_values, degas_values, publish_values,
        apparent_age_values, isochron_values, total_param, isochron_mark,
        sequence,
    ]


def open_240(data: pd.DataFrame, logs01: pd.DataFrame, logs02: pd.DataFrame):
    """
    Open 24.0 version ArArCALC files
    :param data:
    :param logs01:
    :param logs02:
    :return:
    """
    # add default columns for 999, -999, and -1 index
    data[-999] = [0] * data.index.size
    data[999] = [1] * data.index.size
    data[-1] = [np.nan] * data.index.size
    # values index
    sequence_index = [1, 2, ]
    sample_values_index = [16, 17, 21, 22, 26, 27, 31, 32, 36, 37,]
    blank_values_index = [3, 4, 5, 6, 7, 8, 9, 10, 11, 12,]
    corrected_values_index = [
        -999, -999, -999, -999, -999, -999, -999, -999, -999, -999,]
    degas_values_index = [
        # 36 a, c, ca, cl
        136, -999, 137, -999, 139, -999, 138, -999,
        # 37 ca,
        140, -999,
        # 38 cl, a, c, k, ca,
        145, -999, 141, -999, 142, -999, 143, -999, 144, -999,
        # 39 k, ca
        146, -999, 147, -999,
        # 40 r, a, c, k
        148, -999, 149, -999, 150, -999, 151, -999,
    ]
    publish_values_index = [
        99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109,
    ]
    apparent_age_values_index = [
        # f, sf, ages, s, s, s, 39Ar%
        156, 157, 104, 105, -999, -999, 106, 107,
    ]
    isochron_values_index = [
        # normal isochron
        114, 115, 116, 117, 120,
        # inverse
        -1, 125, 126, 127, 128, 131,
        # 39/38 vs 40/38
        -1, -999, -999, -999, -999, -999,
        # 39/40 vs 39/40
        -1, -999, -999, -999, -999, -999,
        # 38/39 vs 40/39
        -1, -999, -999, -999, -999, -999,
        # 3D isochron, 36/39, 38/39, 40/39, r1, r2, r3
        -1, -999, -999, -999, -999, -999, -999, -999, -999, -999,
    ]
    isochron_mark_index = [113,]
    total_param_index = [
        # 0-9
        69, 70, 71, 72, 73, 74, 75, 76, 77, 78,
        # 10-19
        79, 80, 81, 82, 83, 84, 85, 86, 87, 88,
        # 20-29
        89, 90, 91, 92, 93, 94, -999, -999, 63, -999,
        # 30-39
        -999, -999, -1, -1, -999, -999, -999, -999, -999, -999,
        # 40-49
        -999, -999, -999, -999, -999, -999, -999, -999, -999, -999,
        # 50-59
        -999, -999, -999, -999, -999, -999, -999, -999, 65, 49,
        # 60-69
        50, -999, -999, -999, -999, -999, -999, 51, 52, 53,
        # 70-79
        54, -999, -999, -999, -999, -999, -999, -999, -999, -999,
        # 80-89
        -999, -999, -999, -999, -999, -999, -999, -999, -999, -999,
        # 90-99
        -999, -999, -999, -999, -999, -999, -999, -999, -999, -999,
        # 100-99909
        -999, -999, -999, -999, -999, -999, -999, -999, -999, -999,
        # 110-114
        -999, -999, -999, -999, -1,
        # 115-122
        -999, -999, -999, -999, -999, -999, -999, -999,
        # 123-130
        -999, -999, -999, -999, -999, -999, -999, -999,
        # 131-137
        -999, -999, -999, -999, -999, -999, -999,
    ]

    # double transpose to remove keys
    get_data = lambda df, index: pd.concat([df[index].transpose()], ignore_index=True).transpose()

    sample_values = get_data(data, sample_values_index)
    blank_values = get_data(data, blank_values_index)
    corrected_values = get_data(data, corrected_values_index)
    degas_values = get_data(data, degas_values_index)
    publish_values = get_data(data, publish_values_index)
    apparent_age_values = get_data(data, apparent_age_values_index)
    isochron_values = get_data(data, isochron_values_index)
    total_param = get_data(data, total_param_index)
    isochron_mark = get_data(data, isochron_mark_index)
    sequence = get_data(data, sequence_index)

    # irradiation time
    total_param = general_adjustment(
        total_param.copy(), logs02, get_data(data, [59, 58, 57, 60, 61]), data[63][0]
    )
    # Do adjustment
    isochron_mark.replace({4.0: '1'}, inplace=True)
    total_param[34] = logs01[1][36]
    total_param[35] = logs01[1][37]
    total_param[42] = logs01[1][38]
    total_param[43] = logs01[1][39]
    total_param[44] = logs01[1][40]
    total_param[45] = logs01[1][41]

    return [
        sample_values, blank_values, corrected_values, degas_values, publish_values,
        apparent_age_values, isochron_values, total_param, isochron_mark,
        sequence,
    ]


def change_error_type(data: pd.DataFrame, header: pd.Series):
    """
    :param data: 2D DataFrame
    :param header: 1D Series, headers
    :return:
    """
    # 2 sigma to 1 sigma
    tochange = np.flatnonzero(header.where(header.str.contains('2s'), other=False))
    data[tochange.tolist()] = data[tochange.tolist()] / 2

    # percentage errors to absolute errors
    tochange = np.flatnonzero(header.where(header.str.contains('%1s|%2s', regex=True), other=False))
    data[tochange.tolist()] = \
        data[tochange.tolist()] * abs(data[(tochange - 1).tolist()].rename(lambda x: x + 1, axis='columns')) / 100

    return data


def general_adjustment(
        total_param: pd.DataFrame, logs02: pd.DataFrame, experimental_time: pd.DataFrame,
        irradiation_name: str
):
    """
        General handle for all age files, including set irradiation time, initial ratios, error display
    :param total_param:
    :param logs02:
    :param experimental_time:
    :param irradiation_name:
    :return:
    """
    irra_project = pd.Series(logs02[1].copy())
    irradiation_index = np.flatnonzero(irra_project.where(irra_project.str.fullmatch(irradiation_name), other=False))[0]
    irradiation_info = logs02[32][irradiation_index].split('\n')
    # 处理辐照时间
    month_convert = {
        r'^(?i)Jan$': '01', r'^(?i)Feb$': '02', r'^(?i)Mar$': '03', r'^(?i)Apr$': '04', r'^(?i)May$': '05', r'^(?i)Jun$': '06',
        r'^(?i)Jul$': '07', r'^(?i)Aug$': '08', r'^(?i)Sep$': '09', r'^(?i)Oct$': '10', r'^(?i)Nov$': '11', r'^(?i)Dec$': '12'}
    duration_hour = []
    end_time_second = []
    irradiation_end_time = []
    last_time = []
    for each_date in irradiation_info:
        if '/' in each_date:
            _year = each_date.split(' ')[1].split('/')[2]
            _month = each_date.split(' ')[1].split('/')[1].capitalize()
            for pat, val in month_convert.items():
                if re.match(pat, _month):
                    _month = val
                    break
            _day = each_date.split(' ')[1].split('/')[0]
            _hour = each_date.split(' ')[2].split('.')[0]
            _min = each_date.split(' ')[2].split('.')[1]
            end_time_second.append(get_datetime(
                t_year=int(_year), t_month=int(_month), t_day=int(_day),
                t_hour=int(_hour), t_min=int(_min)))
            each_duration_hour = each_date.split(' ')[0].split('.')[0]
            each_duration_min = each_date.split(' ')[0].split('.')[1]
            each_duration = int(each_duration_hour) + round(int(each_duration_min) / 60, 2)
            duration_hour.append(each_duration)
            irradiation_end_time.append(f"{_year}-{_month}-{_day}T{_hour}:{_min}D{each_duration}")
            last_time = f"{_year}-{_month}-{_day}T{_hour}:{_min}"

    experimental_time: pd.DataFrame = experimental_time.apply(pd.to_numeric, errors='ignore', downcast='integer')
    experimental_time: pd.DataFrame = experimental_time.replace(to_replace=month_convert, regex=True)
    total_param[31] = [f"{i[0]}-{i[1]}-{i[2]}T{i[3]}:{i[4]}" for i in experimental_time.values.tolist()]
    total_param[26] = len(irradiation_end_time)
    total_param[27] = 'S'.join(irradiation_end_time)
    total_param[29] = sum(duration_hour)
    total_param[30] = last_time

    stand_time_second = [
        get_datetime(*i) - get_datetime(*re.findall(r'\d+', last_time)) for i in experimental_time.values.tolist()]
    total_param[32] = [i / (3600 * 24 * 365.242) for i in stand_time_second]  # stand year

    # initial ratios & error display settings
    total_param[95] = logs02[9][irradiation_index]  # 38Ar/36Ar air
    total_param[96] = logs02[10][irradiation_index]  # %SD
    total_param[115] = 0
    total_param[116] = 298.56
    total_param[117] = 0.31
    total_param[118] = 298.56
    total_param[119] = 0.31
    total_param[120] = 1
    total_param[121] = 1
    total_param[122] = 1

    return total_param


def open_full_xls(file_path: str, sample_name: str = ''):
    """
    filepath: absolute full path of input file
    return sample instance
    """
    try:
        res = xls.open_xls(file_path)
    except (Exception, BaseException) as e:
        raise 
    start_row = 5
    rows_num = len(res['Sample Parameters']) - 5

    for key, val in res.items():
        res[key] = arr.transpose(val[:start_row + rows_num])
        # 2 sigma errors to 1 sigma
        for i in range(len(res[key])):
            if res[key][i][2] in ['2s', '%2s', '± 2s']:
                _temp = []
                for j in range(rows_num):
                    try:
                        _temp.append(res[key][i][start_row + j] / 2)
                    except TypeError:
                        _temp.append('')
                res[key][i][start_row: start_row + rows_num] = _temp
    if res['Incremental Heating Summary'][13][2] == 'K/Ca':
        _temp, _s = [], []
        for j in range(rows_num):
            try:
                _temp.append(1 / res['Relative Abundances'][13][start_row + j])
                _s.append(err.div((1, 0), (
                    res['Relative Abundances'][13][start_row + j], res['Relative Abundances'][14][start_row + j])))
            except TypeError:
                _temp.append(None)
                _s.append(None)
        res['Incremental Heating Summary'][13][start_row: start_row + rows_num] = _temp
        res['Incremental Heating Summary'][13][start_row: start_row + rows_num] = _s

    # degas: change relative uncertainties to absolute
    for i in range(len(res['Degassing Patterns'])):
        if res['Degassing Patterns'][i][2] in ['%1s', '%2s']:
            _temp = []
            for j in range(rows_num):
                try:
                    _temp.append(res['Degassing Patterns'][i][start_row + j] * res['Degassing Patterns'][i - 1][
                        start_row + j] / 100)
                except TypeError:
                    _temp.append('')
            res['Degassing Patterns'][i][start_row: start_row + rows_num] = _temp
    rows = list(range(start_row, start_row + rows_num))
    sequence_name = arr.partial(res['Procedure Blanks'], rows, 1)
    sequence_value = arr.partial(res['Procedure Blanks'], rows, 2)
    blank_values = arr.partial(
        res['Procedure Blanks'], rows, [3, 4, 5, 6, 7, 8, 9, 10, 11, 12])
    sample_values = arr.partial(
        res['Intercept Values'], rows, [3, 4, 8, 9, 13, 14, 18, 19, 23, 24])
    corrected_values = arr.partial(
        res['Relative Abundances'], rows, [4, 5, 6, 7, 8, 9, 10, 11, 12, 13])
    degas_values = arr.partial(
        res['Degassing Patterns'], rows, [
            4, 5, 6, 7, 8, 9, 10, 11,
            12, 13,
            22, 23, 14, 15, 16, 17, 18, 19, 20, 21,
            24, 25, 26, 27,
            28, 29, 30, 31, 32, 33, 34, 35])
    publish_values = arr.partial(
        res['Incremental Heating Summary'], rows, [4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14])
    apparent_age_values = arr.partial(
        res['Relative Abundances'], rows, [14, 15, 16, 17, -1, -1, 18, 19])
    isochron_values = [[np.nan] * len(rows)] * 39
    isochron_values[0:5] = arr.partial(
        res['Normal Isochron Table'], rows, [4, 5, 6, 7, 10])
    isochron_values[6:11] = arr.partial(
        res['Inverse Isochron Table'], rows, [4, 5, 6, 7, 10])
    total_param = arr.partial(
        res['Sample Parameters'], rows, [
            # 1, 2, -1, # 0-2
            -1, -1, -1, -1,  # 3-6
            -1, -1, -1, -1,  # 7-10
            -1, -1, -1, -1, -1, -1,  # 11-16
            -1, -1, -1, -1,  # 17-20
            -1, -1,  # 21-22
            23, 24, -1, -1, -1, -1,  # 23-28
            -1, -1,  # 29-30
            22, -1, -1, -1,  # 31-34
            -1, -1,  # 35-36
            -1, -1,  # 37-38
            -1, -1,  # 39-40
            -1, -1,  # 41-42
            -1, -1,  # 43-44
            -1, -1,  # 45-46
            -1, -1,  # 47-48
            -1, -1,  # 49-50
            -1, -1,  # 51-52
            -1, -1,  # 53-54
            -1, -1,  # 55-56
            -1, -1,  # 57-58
            -1, -1,  # 59-60
            -1, -1, -1, -1, -1, -1, -1, -1, -1,  # 61-69
            10, 11, 12, 13,  # 70-73
            -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,  # 74-85
            -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,  # 86-95
            -1, -1, -1, -1,  # 96-99
            -1, -1, -1, -1,  # 100-103
            -1, -1, -1, -1,  # 104-107
            -1, -1, -1, -1,  # 108-111
            -1, -1, -1, -1,  # 112-115
            -1, -1, -1, -1, -1, -1, -1, -1,  # 116-123
            -1, -1, -1, -1, -1, -1, -1, -1,  # 124-131
            -1, -1, -1, -1, -1,  # 132-136
            -1, -1, -1, -1,  # 137-140
        ])

    month_convert = {'Jan': '01', 'Feb': '02', 'Mar': '03', 'Apr': '04', 'May': '05', 'Jun': '06',
                     'Jul': '07', 'Aug': '08', 'Sep': '09', 'Oct': '10', 'Nov': '11', 'Dec': '12'}
    experiment_start_time = []
    for i in arr.transpose(
            arr.partial(res['Sample Parameters'], rows, [18, 17, 16, 19, 20])):
        _year = str(i[0]) if '.' not in str(i[0]) else str(i[0]).split('.')[0]
        _month = str(month_convert[str(i[1]).capitalize()]) if str(i[1]).capitalize() in month_convert.keys() else str(
            i[1]).capitalize()
        _day = str(i[2]) if '.' not in str(i[2]) else str(i[2]).split('.')[0]
        _hour = str(i[3]) if '.' not in str(i[3]) else str(i[3]).split('.')[0]
        _min = str(i[4]) if '.' not in str(i[4]) else str(i[4]).split('.')[0]
        experiment_start_time.append(f"{_year}-{_month}-{_day}T{_hour}:{_min}")

    total_param[31] = experiment_start_time
    total_param[0:26] = arr.partial(
        res['Irradiation Constants'], rows,
        [3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28])

    isochron_values[11] = [''] * len(isochron_values[0])
    isochron_values[17] = [''] * len(isochron_values[0])
    isochron_values[23] = [''] * len(isochron_values[0])
    isochron_values[12:17] = isochron.get_data(
        degas_values[20], degas_values[21], degas_values[24], degas_values[25], degas_values[10], degas_values[11]
    )  # 39/38, 40/38
    isochron_values[18:23] = isochron.get_data(
        degas_values[20], degas_values[21], degas_values[10], degas_values[11], degas_values[24], degas_values[25]
    )  # 39/40, 38/40
    isochron_values[24:29] = isochron.get_data(
        degas_values[10], degas_values[11], degas_values[24], degas_values[25], degas_values[20], degas_values[21]
    )  # 38/39, 40/39
    isochron_mark = ['1' if i == 'P' else '' for i in arr.partial(
        res['Normal Isochron Table'], rows, 3)]

    sample_info = {
        'sample': {'name': sample_name, 'material': 'MATERIAL', 'location': 'LOCATION'},
        'researcher': {'name': 'RESEARCHER', 'addr': 'ADDRESS'},
        'laboratory': {'name': 'LABORATORY', 'addr': 'ADDRESS', 'analyst': 'ANALYST'}
    }

    res = [sample_values, blank_values, corrected_values, degas_values, publish_values,
           apparent_age_values, isochron_values, total_param, [isochron_mark],
           [sequence_name, sequence_value]]

    return dict(zip(['smp', 'blk', 'cor', 'deg', 'pub', 'age', 'iso', 'pam', 'mak', 'seq'], res)), sample_info


class ArArCalcFile:
    def __init__(self, file_path: str, sample_name: str = ""):
        self.file_path = file_path
        self.sample_name = sample_name
        self.supported_versions = ['25.2', '24.0']
        self.content = pd.DataFrame([])
        self.sample_info = {
            'sample': {'name': 'undefined', 'material': 'undefined',},
            'researcher': {'name': 'undefined'},
            'laboratory': {'name': 'undefined', 'analyst': 'undefined', 'info': 'undefined',}
        }

    def open(self):
        book_contents = read_calc_file(self.file_path)
        # create data frames for book values
        content = pd.DataFrame(book_contents['Data Tables'])
        logs01 = pd.DataFrame(book_contents.get('Logs01', [[]]))
        logs02 = pd.DataFrame(book_contents.get('Logs02', [[]]))
        logs03 = pd.DataFrame(book_contents.get('Logs03', [[]]))

        start_row = 5
        sequence_num = int(logs03[2][0])
        header = content.loc[2]
        data = pd.concat([content.loc[list(range(start_row, start_row + sequence_num))]], ignore_index=True).apply(
            pd.to_numeric, errors='ignore'
        )

        # check version
        version = logs03[3][0]
        if version == '25.2':
            handler = open_252
            material = data[45][0]
            analyst = data[47][0]
        elif version == '24.0':
            handler = open_240
            material = data[46][0]
            analyst = data[48][0]
        else:
            raise ValueError(f'Unsupported version: {version}')

        # change error type, 2sigma to 1sigma..., relative errors to absolute errors
        data = change_error_type(data, header)
        # get full data frames
        # ['smp', 'blk', 'cor', 'deg', 'pub', 'age', 'iso', 'pam', 'inf', 'mak', 'seq',] are abbreviations for
        #     [sample_values, blank_values, corrected_values, degas_values, publish_values,
        #      apparent_age_values, isochron_values, total_param, sample_info, isochron_mark,
        #      sequence_name, sequence_value]
        self.content = pd.concat(
            handler(data.copy(), logs01, logs02), axis=1,
            keys=['smp', 'blk', 'cor', 'deg', 'pub', 'age', 'iso', 'pam', 'mak', 'seq', ],
        )

        # set error format for parameters, change to percentage
        # list(range(1, 26, 2)) irradiation correction constants
        # list(range(39, 58, 2)) decay constants
        # list(range(68, 97, 2)) J, MDF, other constants
        for column in list(range(1, 26, 2)) + list(range(68, 71, 2)):  # irradiation constants and J, MDF uncertainties
            self.content.loc[:, ('pam', column)] = \
                round(self.content['pam', column].astype("float") / (
                    self.content['pam', column - 1].astype("float")) * 100, 6).replace(np.nan, 0)

        # sample info
        self.sample_info = {
            'sample': {
                'name': self.sample_name or data[44][0],
                'material': material, 'location': 'LOCATION'
            },
            'researcher': {'name': data[64][0]},
            'laboratory': {
                'name': logs01[1][44],
                'analyst': analyst,
                'info': '\n'.join([logs01[1][45], logs01[1][46], logs01[1][47]]),
            }
        }

        return self

    def get_content(self):
        return self.content

    def get_smp_info(self):
        return self.sample_info

    def get_df(self):
        return self.content

    def get_list(self):
        return self.content



#  Copyright (C) 2024 Yang. - All Rights Reserved
"""
# ==========================================
# Copyright 2024 Yang
# ararpy - test.py
# ==========================================
#
#
#
"""
import ararpy as ap
import os


def test():
    example_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), r'examples')
    print(f"Running: ararpy.test()")
    print(f"============= Open an example .arr file =============")
    file_path = os.path.join(example_dir, r'22WHA0433.arr')
    sample = ap.from_arr(file_path=file_path)
    # file_path = os.path.join(example_dir, r'22WHA0433.age')
    # sample = ap.from_age(file_path=file_path)
    print(f"{file_path = }")
    print(f"sample = from_arr(file_path=file_path)")
    print(f"{sample.name() = }")
    sample.name("new name")
    print(f"{sample.name() = }")
    print(f"{sample.help() = }")
    print(f"{sample.parameters() = }")
    print(f"{sample.parameters().to_df() = }")
    print(sample.show_data())
    print(sample.sample())
    print(sample.blank().to_df().iloc[:, [1, 2, 3]])


def export_pdf_demo():
    import numpy as np
    import pdf_maker as pm

    example_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), r'examples')

    # ------ 将以下五个样品的年龄谱图组合到一起 -------
    arr_files = [
        os.path.join(example_dir, r'22WHA0433.arr'),
        os.path.join(example_dir, r'20WHA0103.age'),
    ]
    colors = ['#1f3c40', '#e35000', '#e1ae0f', '#3d8ebf']
    series = []

    # ------ 构建数据 -------
    for index, file in enumerate(arr_files):
        smp = ap.from_arr(file_path=file) if file.endswith('.arr') else ap.from_age(file_path=file)
        age = smp.ApparentAgeValues[2:4]
        ar = smp.DegasValues[20]
        data = ap.calc.spectra.get_data(*age, ar, cumulative=False)
        series.append({
            'type': 'series.line', 'id': f'line{index * 2 + 0}', 'name': f'line{index * 2 + 0}', 'color': colors[index],
            'data': np.transpose([data[0], data[1]]).tolist(), 'line_caps': 'square',
        })
        series.append({
            'type': 'series.line', 'id': f'line{index * 2 + 1}', 'name': f'line{index * 2 + 1}', 'color': colors[index],
            'data': np.transpose([data[0], data[2]]).tolist(), 'line_caps': 'square',
        })
        series.append({
            'type': 'text', 'id': f'text{index * 2 + 0}', 'name': f'text{index * 2 + 0}', 'color': colors[index],
            'text': f'{smp.name()}<r>{round(smp.Info.results.age_plateau[0]["age"], 2)}',
            'size': 10, 'data': [[index * 15 + 5, 23]],
        })
    data = {
        "data": [
            {
                'xAxis': [{'extent': [0, 100], 'interval': [0, 20, 40, 60, 80, 100],
                           'title': 'Cumulative <sup>39</sup>Ar Released (%)', 'nameLocation': 'middle', }],
                'yAxis': [{'extent': [0, 250], 'interval': [0, 50, 100, 150, 200, 250],
                           'title': 'Apparent Age (Ma)', 'nameLocation': 'middle', }],
                'series': series
            }
        ],
        "file_name": "WHA",
        "plot_names": ["all age plateaus"],
    }

    # write pdf
    file = pm.NewPDF(filepath=os.path.join(example_dir, f"{data['file_name']}.pdf"))
    for index, each in enumerate(data['data']):
        # rich text tags should follow this priority: color > script > break
        file.text(page=index, x=50, y=780, line_space=1.2, size=12, base=0, h_align="left",
                  text=f"The PDF can be edited with Adobe Acrobat, Illustrator and CorelDRAW")
        cv = ap.smp.export.export_chart_to_pdf(each)
        file.canvas(page=index, base=0, margin_top=5, canvas=cv, unit="cm", h_align="middle")
        if index + 1 < len(data['data']):
            file.add_page()

    # save pdf
    file.save()


if __name__ == "__main__":
    test()
    # export_pdf_demo()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2023 Yang
# ararpy - smp - initial
# ==========================================
#
#
#
"""
import os
import pickle
import uuid
import pandas as pd
import numpy as np
import copy
from typing import List, Union, Optional
from ..calc import arr, err
from ..files import calc_file
from . import (sample as samples, basic, table, raw as smp_raw)

Sample = samples.Sample
Table = samples.Table
Plot = samples.Plot
RawData = samples.RawData
Sequence = samples.Sequence
ArArBasic = samples.ArArBasic

plateau_res_keys = [
    'F', 'sF', 'Num', 'MSWD', 'Chisq', 'Pvalue', 'age', 's1', 's2', 's3', 'Ar39',
    'rs',  # 'rs' means relative error of the total sum
]
PLATEAU_RES = dict(zip(plateau_res_keys, [np.nan for i in plateau_res_keys]))
iso_res_keys = [
    'k', 'sk', 'm1', 'sm1',
    'MSWD', 'abs_conv', 'iter', 'mag', 'R2', 'Chisq', 'Pvalue',
    'rs',  # 'rs' means relative error of the total sum
    'age', 's1', 's2', 's3',
    'conv', 'initial', 'sinitial', 'F', 'sF',
]
ISO_RES = dict(zip(
    iso_res_keys, [np.nan for i in iso_res_keys])
)
spectra_res_keys = [
    'F', 'sF', 'Num', 'MSWD', 'Chisq', 'Pvalue', 'age', 's1', 's2', 's3', 'Ar39',
    'rs',  # 'rs' means relative error of the total sum
]
SPECTRA_RES = dict(zip(
    spectra_res_keys, [np.nan for i in spectra_res_keys])
)
preference_keys = [
    'decimalPlaces', 'toPrecision', 'confidenceLevel', 'ageUnit'
]
PREFERENCE_RES = dict(zip(
    preference_keys, [6, 'fixed', 1, 'Ma'])
)


# create sample instance
def create_sample_from_df(content: pd.DataFrame, smp_info: dict):
    """

    Parameters
    ----------
    content : [
            sample_values, blank_values, corrected_values, degas_values, publish_values,
            apparent_age_values, isochron_values, total_param, sample_info, isochron_mark,
            sequence_name, sequence_value
        ]
    smp_info : dict

    Returns
    -------
    Sample instance
    """
    content_dict = content.to_dict('list')
    res = dict(zip([key[0] for key in content_dict.keys()], arr.create_arr((len(content_dict), 0))))
    for key, val in content_dict.items():
        res[key[0]] = res[key[0]] + [val]

    return create_sample_from_dict(content=res, smp_info=smp_info)


def create_sample_from_dict(content: dict, smp_info: dict):
    """
    content:
        {
            'smp': [], 'blk': [], 'cor': [], 'deg': [], 'pub': [],
            'age': [], 'iso': [], 'pam': [], 'mak': [], 'seq': [],
            'seq': []
        }
    return sample instance
    """
    # Create sample file
    smp = Sample()
    # Initializing
    initial(smp)
    smp.SampleIntercept = content['smp']
    smp.BlankIntercept = content['blk']
    smp.CorrectedValues = content['cor']
    smp.DegasValues = content['deg']
    smp.PublishValues = content['pub']
    smp.ApparentAgeValues = content['age']
    smp.IsochronValues = content['iso']
    smp.TotalParam = content['pam']
    smp.IsochronMark = content['mak'][0]
    smp.SequenceName = content['seq'][0]
    smp.SequenceValue = content['seq'][1]

    smp.Info = basic.update_plot_from_dict(smp.Info, smp_info)

    smp.SelectedSequence1 = [index for index, item in enumerate(smp.IsochronMark) if item == 1 or item == '1']
    smp.SelectedSequence2 = [index for index, item in enumerate(smp.IsochronMark) if item == 2 or item == '2']
    smp.UnselectedSequence = [index for index, item in enumerate(smp.IsochronMark) if item not in [1, 2, '1', '2']]
    #
    smp.Info.results.selection[0]['data'] = smp.SelectedSequence1
    smp.Info.results.selection[1]['data'] = smp.SelectedSequence2
    smp.Info.results.selection[2]['data'] = smp.UnselectedSequence

    return smp


def initial(smp: Sample):
    # 已更新 2023/7/4
    smp.TotalParam = arr.create_arr((len(samples.TOTAL_PARAMS_HEADERS) - 3, 0))
    smp.BlankIntercept = arr.create_arr((len(samples.BLANK_INTERCEPT_HEADERS) - 3, 0))
    smp.SampleIntercept = arr.create_arr((len(samples.SAMPLE_INTERCEPT_HEADERS) - 3, 0))
    smp.PublishValues = arr.create_arr((len(samples.PUBLISH_TABLE_HEADERS) - 3, 0))
    smp.DecayCorrected = arr.create_arr((10, 0))
    smp.CorrectedValues = arr.create_arr((len(samples.CORRECTED_HEADERS) - 3, 0))
    smp.DegasValues = arr.create_arr((len(samples.DEGAS_HEADERS) - 3, 0))
    smp.ApparentAgeValues = arr.create_arr((len(samples.SPECTRUM_TABLE_HEADERS) - 3, 0))
    smp.IsochronValues = arr.create_arr((len(samples.ISOCHRON_TABLE_HEADERS) - 3, 0))

    # Doi
    if not hasattr(smp, 'Doi') or getattr(smp, 'Doi') in (None, ""):
        setattr(smp, 'Doi', str(uuid.uuid4().hex))

    # Info
    setattr(smp, 'Info', ArArBasic(
        id='0', name='info', attr_name='Info', arr_version=samples.VERSION,
        sample=ArArBasic(
            name='SAMPLE NAME', material='MATERIAL', location='LOCATION', type='Unknown', method='',
            sequence_unit='', weight=''
        ),
        researcher=ArArBasic(
            name='RESEARCHER', addr='ADDRESS', email='EMAIL'
        ),
        laboratory=ArArBasic(
            name='LABORATORY', addr='ADDRESS', email='EMAIL', info='INFORMATION', analyst='ANALYST'
        ),
        results=ArArBasic(
            name='RESULTS', plateau_F=[], plateau_age=[], total_F=[], total_age=[],
            isochron_F=[], isochron_age=[], J=[],
            # set1=result_set_1, set2=result_set_2,
            isochron={
                'figure_2': {
                    0: copy.deepcopy(ISO_RES), 1: copy.deepcopy(ISO_RES), 2: copy.deepcopy(ISO_RES)},
                'figure_3': {
                    0: copy.deepcopy(ISO_RES), 1: copy.deepcopy(ISO_RES), 2: copy.deepcopy(ISO_RES)},
                'figure_4': {
                    0: copy.deepcopy(ISO_RES), 1: copy.deepcopy(ISO_RES), 2: copy.deepcopy(ISO_RES)},
                'figure_5': {
                    0: copy.deepcopy(ISO_RES), 1: copy.deepcopy(ISO_RES), 2: copy.deepcopy(ISO_RES)},
                'figure_6': {
                    0: copy.deepcopy(ISO_RES), 1: copy.deepcopy(ISO_RES), 2: copy.deepcopy(ISO_RES)},
                'figure_7': {
                    0: copy.deepcopy(ISO_RES), 1: copy.deepcopy(ISO_RES), 2: copy.deepcopy(ISO_RES)},
            },
            age_plateau={
                0: copy.deepcopy(PLATEAU_RES), 1: copy.deepcopy(PLATEAU_RES), 2: copy.deepcopy(PLATEAU_RES)},
            age_spectra={
                'TGA': copy.deepcopy(SPECTRA_RES),
                0: copy.deepcopy(SPECTRA_RES), 1: copy.deepcopy(SPECTRA_RES), 2: copy.deepcopy(SPECTRA_RES),
            },
            selection={
                0: {'data': [], 'name': 'set1'},
                1: {'data': [], 'name': 'set2'},
                2: {'data': [], 'name': 'set3'}
            }
        ),
        reference=ArArBasic(
            name='REFERENCE', journal='JOURNAL', doi='DOI'
        ),
        preference=copy.deepcopy(PREFERENCE_RES),
        experiment=ArArBasic(
            name='', mass_spec='', collectors='', step_num=0
        ),
    ))

    decimal_places = PREFERENCE_RES['decimalPlaces']
    # Plots and Tables
    setattr(smp, 'UnknownTable', Table(
        id='1', name='Unknown', header=samples.SAMPLE_INTERCEPT_HEADERS, decimal_places=decimal_places,
        text_indexes=[0, 1, 2],
        # numeric_indexes=list(range(1, 20))
    ))
    setattr(smp, 'BlankTable', Table(
        id='2', name='Blank', header=samples.BLANK_INTERCEPT_HEADERS, decimal_places=decimal_places,
        text_indexes=[0, 1, 2],
        # numeric_indexes=list(range(1, 20))
    ))
    setattr(smp, 'CorrectedTable', Table(
        id='3', name='Corrected', header=samples.CORRECTED_HEADERS, decimal_places=decimal_places,
        text_indexes=[0, 1, 2],
        # numeric_indexes=list(range(1, 35))
    ))
    setattr(smp, 'DegasPatternTable', Table(
        id='4', name='Degas Pattern', header=samples.DEGAS_HEADERS, decimal_places=decimal_places,
        text_indexes=[0, 1, 2],
        # numeric_indexes=list(range(1, 35))
    ))
    setattr(smp, 'PublishTable', Table(
        id='5', name='Publish', header=samples.PUBLISH_TABLE_HEADERS, decimal_places=decimal_places,
        text_indexes=[0, 1, 2],
        # numeric_indexes=list(range(1, 20))
    ))
    setattr(smp, 'AgeSpectraTable', Table(
        id='6', name='Age Spectra', header=samples.SPECTRUM_TABLE_HEADERS, decimal_places=decimal_places,
        text_indexes=[0, 1, 2],
        # numeric_indexes=list(range(1, 26))
    ))
    setattr(smp, 'IsochronsTable', Table(
        id='7', name='Isochrons', header=samples.ISOCHRON_TABLE_HEADERS, decimal_places=decimal_places,
        text_indexes=[0, 1, 2, 8, 14, 20, 26, 32],
        # numeric_indexes=[1, *list(range(3, 42))]
    ))
    setattr(smp, 'TotalParamsTable', Table(
        id='8', name='Total Params', header=samples.TOTAL_PARAMS_HEADERS, decimal_places=decimal_places,
        text_indexes=[0, 1, 2, 29, 30, 32, 33, 60, 99, *list(range(100, 115))],
        # numeric_indexes=list(range(1, 120)),
    ))

    initial_plot_styles(smp)

    return smp


def initial_plot_styles(smp: Sample, except_attrs=None):
    """
    Initialize plot components styles based on Default Styles. Except attrs is a list containing attrs
    that are not expected to be initialized.
    Judgment order:
        1. The attr name is in except attrs and the sample has this attr: skip
        2. The value is not a dict instance: setattr()
        3. The sample has attr and it is a Set/Label/Text/Axis instance: iteration
    """

    if except_attrs is None:
        except_attrs = []

    def set_attr(obj, name, value):
        if name in except_attrs and hasattr(obj, name):
            pass
        elif not isinstance(value, dict):
            setattr(obj, name, value)
        else:
            if not (hasattr(obj, name) and isinstance(getattr(obj, name), Plot.BasicAttr)):
                setattr(obj, name, getattr(Plot, value['type'].capitalize())())
            for k, v in value.items():
                set_attr(getattr(obj, name), k, v)

    default_styles = get_default_plot_style(smp)
    for figure_index, figure_attr in default_styles.items():
        plot = getattr(smp, figure_attr['attr_name'], Plot())
        for key, attr in figure_attr.items():
            set_attr(plot, key, attr)


def get_default_plot_style(smp: Sample):
    sample_type = smp.Info.sample.type
    try:
        age_unit = str(smp.Info.preference['ageUnit']).capitalize()
    except:
        age_unit = "Undefined"
    return copy.deepcopy(samples.DEFAULT_PLOT_STYLES(sample_type, age_unit))


def re_set_smp(smp: Sample):
    std = initial(Sample())
    basic.get_merged_smp(smp, std)
    return check_version(smp)


def check_version(smp: Sample):
    """

    Parameters
    ----------
    smp

    Returns
    -------

    """
    if smp.version != samples.VERSION:
        std = initial(Sample())
        basic.get_merged_smp(smp, std)

    try:
        version = int(smp.version)
    except ValueError:
        return smp

    # 20250328:  # Experiment info
    smp.Info.experiment.name = smp.name()
    smp.Info.experiment.step_num = smp.sequence().size

    # old version: add masses and gain factors
    if version < 20240730:
        gains = np.ones([10, smp.Info.experiment.step_num])
        gains[[1, 3, 5, 7, 9], :] = 0
        smp.TotalParam[126:136] = gains.tolist()
        gains[0] = 35.96754628
        gains[2] = 36.9667759
        gains[4] = 37.9627322
        gains[6] = 38.964313
        gains[8] = 39.962383123
        smp.TotalParam[71:81] = gains.tolist()
        smp.version = "20240730"

    # 20250404:  # Normalization for steps with different J values
    doNormalize = True
    v, sv = [], []
    try:
        if smp.Info.sample.type.lower() == "unknown":
            v, sv = smp.TotalParam[67:69]
            sv = np.multiply(v, sv) / 100
        elif smp.Info.sample.type.lower() == "air":
            v, sv = smp.TotalParam[93:95]
            sv = np.multiply(v, sv) / 100
        elif smp.Info.sample.type.lower() == "standard":
            v, sv = smp.TotalParam[59:61]
    except:
        doNormalize = False
    smp.TotalParam[136:138] = [
        [1 if v[0] == each or not doNormalize else v[0] / each for each in v],
        [0 if v[0] == v[i] or not doNormalize else err.div((v[0], sv[0]), (v[i], sv[i])) for i in range(len(v))]
    ]

    # 20251001:  # add marks for every table
    smp = smp.recalculate(re_set_table=True, re_table_style=True)
    return smp


# create
def from_empty(file_path: str = '', sample_name: str = None):
    """
    Parameters
    ----------
    file_path
    sample_name

    Returns
    -------

    """
    sample = Sample()
    # initial settings
    initial(sample)
    if sample_name is not None:
        sample.Info.sample.name = sample_name
    return sample


# create
def from_arr_files(file_path, sample_name: str = ""):
    """
    file_path: full path of input file
    name： samplename
    return sample instance
    """

    class RenameUnpickler(pickle.Unpickler):
        def find_class(self, module: str, name: str):
            SAMPLE_MODULE = Sample().__module__
            renamed_module = module
            if '.sample' in module and module != SAMPLE_MODULE:
                renamed_module = SAMPLE_MODULE
            try:
                return super(RenameUnpickler, self).find_class(renamed_module, name)
            except AttributeError:
                return super(RenameUnpickler, self).find_class(renamed_module, 'ArArBasic')

    def renamed_load(file_obj):
        return RenameUnpickler(file_obj).load()

    with open(file_path, 'rb') as f:
        sample = renamed_load(f)
    # Check arr version
    # recalculation will not be applied automatically
    return check_version(sample)


# create
def from_calc_files(file_path: str, **kwargs):
    """

    Parameters
    ----------
    file_path

    Returns
    -------

    """
    file = calc_file.ArArCalcFile(file_path=file_path, **kwargs).open()
    sample = create_sample_from_df(file.get_content(), file.get_smp_info())
    return check_version(sample)


# create
def from_full_files(file_path: str, sample_name: str = None):
    """
    Parameters
    ----------
    file_path
    sample_name

    Returns
    -------

    """
    if sample_name is None:
        sample_name = str(os.path.split(file_path)[-1]).split('.')[0]
    content, sample_info = calc_file.open_full_xls(file_path, sample_name)
    sample = create_sample_from_dict(content=content, smp_info=sample_info)
    return check_version(sample)


# create
def from_raw_files(file_path: Union[str, List[str]], input_filter_path: Union[str, List[str]],
                   mapping: Optional[List[dict]] = None) -> Sample:
    raw = smp_raw.to_raw(file_path, input_filter_path)
    raw.do_regression()
    return from_raw_data(raw, mapping)


def from_raw_data(raw: RawData, mapping: Optional[List[dict]] = None) -> Sample:
    """
    Parameters
    ----------
    raw
    mapping :
        mapping is a list of dictionaries with two keys of blank and unknown,
        for example, mapping = [
            {'blank': blank_name, 'unknown': unknown_name_1},
            ...,
            {'blank': blank_name, 'unknown': unknown_name_1}
        ]

    Returns
    -------

    """
    if mapping is None:
        mapping = []
        _b: Sequence = raw.get_sequence(index=True, flag='is_blank', unique=True)
        for _index, _seq in enumerate(raw.sequence):
            if _seq.is_blank():
                _b = _seq
                continue
            else:
                mapping.append({'unknown': _seq.name, 'blank': _b.name})

    # 创建sample
    sample = Sample()
    sample.RawData = raw
    initial(sample)
    unknown_intercept, blank_intercept = [], []
    for row in mapping:
        row_unknown_intercept = []
        row_blank_intercept = []

        unknown: Sequence = raw.get_sequence(row['unknown'], flag='name')
        if row['blank'].lower() == "Interpolated Blank".lower():
            blank: Sequence = arr.filter(
                raw.interpolated_blank, func=lambda seq: seq.datetime == unknown.datetime,
                get=None, unique=True)
        else:
            blank: Sequence = raw.get_sequence(row['blank'], flag='name')
        for i in range(5):
            row_unknown_intercept = arr.multi_append(row_unknown_intercept, *unknown.results[i][int(unknown.fitting_method[i])][:2])
            row_blank_intercept = arr.multi_append(row_blank_intercept, *blank.results[i][int(blank.fitting_method[i])][:2])

        unknown_intercept.append(row_unknown_intercept)
        blank_intercept.append(row_blank_intercept)
        sample.SequenceName.append(unknown.name)
        sample.SequenceValue.append('')
        sample.TotalParam[31].append(unknown.datetime)

    sample.SampleIntercept = arr.transpose(unknown_intercept)
    sample.BlankIntercept = arr.transpose(blank_intercept)
    sample.UnselectedSequence = list(range(len(sample.SequenceName)))
    sample.SelectedSequence1 = []
    sample.SelectedSequence2 = []
    #
    sample.Info.results.selection[0]['data'] = sample.SelectedSequence1
    sample.Info.results.selection[1]['data'] = sample.SelectedSequence2
    sample.Info.results.selection[2]['data'] = sample.UnselectedSequence

    table.update_table_data(sample)  # Update table after submission row data and calculation

    # sample.TotalParam[31] = [raw.get_sequence(row['unknown'], flag='name').datetime for row in mapping]

    return check_version(sample)
